defstar {
  name {AkObservation}
  domain {DE}
  author {G. C. Ewing}
  location {Akaroa library}
  desc {
Submits observations to Akaroa and outputs final simulation results.
}
  explanation {
.EQ
delim off
.EN
This star provides an interface to Akaroa so that a Ptolemy
simulation may be used as a simulation engine under Akaroa.
.pp
A simulation should contain exactly ONE instance of this
star. It accepts any number of inputs and treats them as 
parameters to be observed. Akaroa will collect and analyse 
observations until it decides that sufficient precision has
been reached. Then it will produce a message containing the
final results and halt the simulation.
.pp
When this star is present in a simulation, the Ptolemy random
number source is replaced with one which obtains random
numbers from Akaroa, so all of the random distribution
generators supplied with Ptolemy may be used normally.
However, you will not be able to change the random number
seed using the Pigi "set-seed" command.
.pp
When running the simulation stand-alone under Pigi, you
should enter a very large time for "When to stop"
in the "Run" dialogue box, to ensure that enough observations
are produced to satisfy Akaroa. You should get an alert
box informing you that "Akaroa is stopping the simulation."
If the simulation halts but you do not get this confirmation,
you have set the stopping time too soon.
.pp
The results are output as a message of type "AkResultMessage".
When fed into a TclScript star, this message appears as a
Tcl list with an element for each parameter, where each element
is a list of six items:
.pp
{count trans mean variance delta conf}
.pp
The companion star "ShowAkaroaResults" may be used to display
the results and as an example of how to manipulate the results
using a Tcl script.
.pp
For more information about the AkResultMessage message, see
the "AkResultMessage.H" header file.
.pp
NOTE: After producing the result message, the next time the
Akaroa star is fired it will halt the simulation. Therefore,
any processing of the results by other stars and galaxies must
be accomplished without any delay (or at least before the
next input event to the Akaroa star is scheduled).
.pp
To run a Ptolemy simulation on multiple hosts under Akaroa,
compile the simulation into a Tcl script, and give akrun
a command which will invoke the script via ptcl.
  }
  
  hinclude {
    "ACG.h"
    "ak_begin.H"
    "akaroa.H"
    "akaroa/engine/observation_analyser.H"
    "akaroa/engine/ak_reporter.H"
    "environment.H"
    "ak_end.H"
  }

  ccinclude {
    "iostream.h"
    "strstream.h"
    "SimControl.h"
    "Matrix.h"
    "AkResultMessage.H"
    "akaroa/result.H"
  }

  inmulti {
    name {input}
    type {float}
  }

  output {
    name {output}
    type {message}
  }

  state {
    name {precision}
    type {float}
    default {0.05}
  }

  state {
    name {confidence}
    type {float}
    default {0.95}
  }

  state {
    name {debug}
    type {int}
    default {FALSE}
  }

  protected {
    ObservationAnalyser *oa;
    AkReporter *reporter;
    int count;
    ACG *ak_rng;
    int stopFlag;
    double stopTime;
  }

  code {
    extern ACG *gen;

    AkResultMessage::AkResultMessage(int n) {
      numParams = n;
      params = new AkResult[n];
    }

    AkResultMessage::~AkResultMessage() {
      delete[] params;
    }

    const AkResult &AkResultMessage::param(int i) const {
      return params[i-1];
    }

    AkResult &AkResultMessage::param(int i) {
      return params[i-1];
    }

    StringList AkResultMessage::print() const {
      strstream str;
      for (int i = 1; i <= numParams; i++) {
	const AkResult &r = param(i);
	str << "{" 
	    << r.count << " " 
	    << r.trans << " "
	    << r.mean << " "
	    << r.variance << " "
	    << r.delta << " "
	    << r.conf
	    << "} ";
      }
      str << ends;
      StringList sl(str.str());
      str.freeze(0);
      return sl;
    }

    Message *AkResultMessage::clone() const {
      AkResultMessage *copy = new AkResultMessage(numParams);
      for (int i = 1; i <= numParams; i++)
	copy->param(i) = param(i);
      return copy;
    }
	
    const char *AkResultMessage::dataType() {return "AkResultMessage";}

    ISA_FUNC(AkResultMessage, Message);
      
    class PtolemyReporter: public AkReporter {
      AkResultMessage *msg;
      DEAkObservation *star;

    public:
      PtolemyReporter(DEAkObservation *s) {
	star = s;
	msg = 0;
      }

      virtual void Begin(int n) {
	msg = new AkResultMessage(n);
      }

      virtual void Report(int n, const AkResult &r) {
	msg->param(n) = r;
      }

      virtual void End() {
	star->completionTime = star->arrivalTime;
	star->output.put(star->completionTime) << *msg;
	msg = 0;
      }
    };

    class AkRNG : public ACG {

      long spareBits;
      int numSpareBits;

    public:

      AkRNG() {
	numSpareBits = 0;
      }

      virtual _G_uint32_t asLong() {
	if (numSpareBits == 0) {
	  spareBits = AkRandom();
	  numSpareBits = 31;
	}
	long bits = (AkRandom() << 1) | (spareBits & 1);
	spareBits >>= 1;
	--numSpareBits;
	return bits ^ (bits << 1);
      }

    };
  }

  constructor {
    oa = 0;
    reporter = new PtolemyReporter(this);
    ak_rng = new AkRNG;
    stopTime = 0;
  }

  destructor {
    delete oa;
    delete reporter;
  }

  setup {
    if (oa) {
      delete oa;
      oa = 0;
    }
    count = 0;
    if (gen != ak_rng) {
      if (debug)
	Error::message("Installing Akaroa random number generator");
      gen = ak_rng;
    }
    else {
      if (debug)
	Error::message("Akaroa random number generator installed");
    }
    Environment *env = DefaultEnvironment();
    env->SetReal("Precision", precision);
    env->SetReal("Confidence", confidence);
    stopFlag = FALSE;
  }

  go {
    if (stopFlag) {
      if (arrivalTime > stopTime)
	SimControl::requestHalt();
      return;
    }
    if (!oa) {
      int numParameters = input.numberPorts();
      if (debug) {
	char buf[200];
	sprintf(buf, 
		"Creating an ObservationAnalyser for %ld parameters"
		" with reporter %p",
		numParameters, reporter); 
	Error::message(buf);
      }
      oa = new ObservationAnalyser(numParameters, reporter);
    }
    MPHIter nextport(input);
    InDEPort *port;
    int i = 1;
    while (port = (InDEPort*)(nextport++)) {
      int paramNumber = i++;
      if (port->dataNew) {
	double value = port->get();
	if (debug) {
	  ostrstream str;
	  str << "*AkObservation: Param " << paramNumber
	       << ", observation " << ++count << ": " << value << endl;
	  Error::message(str.str());
	  str.freeze(0);
	}
	oa->ProcessObservation(paramNumber, value);
      }
    }
    if (oa->SimulationOver()) {
      oa->Report();
      Error::message("Akaroa is stopping the simulation.");
      stopFlag = TRUE;
      stopTime = arrivalTime;
    }
  }
}
